"""Unit tests for the Ray cluster component"""
from unittest import mock
from ray_cluster.component import ray_cluster

ray_comp_func = ray_cluster.python_func


@mock.patch("time.sleep")
@mock.patch("ray.job_submission.JobSubmissionClient")
@mock.patch("ray_cluster.utils.utils.RayCluster")
def test_ray_cluster_comp_head_node(
    mock_ray_cluster,
    mock_job_submission_client,
    mock_sleep,
):
    """Test the component functionality in the head node"""
    mock_sleep.return_value = "1s"
    mock_ray_cluster.return_value = mock.MagicMock(workerpool_type="workerpool0")
    ray_comp_func(ray_task_cmd="test command")
    assert mock_ray_cluster.return_value.start_ray_cluster.called
    assert mock_ray_cluster.return_value.wait_for_nodes.called
    assert mock_job_submission_client.return_value.submit_job.called
    assert mock_job_submission_client.return_value.get_job_info.called


@mock.patch("time.sleep")
@mock.patch("ray_cluster.utils.utils.RayCluster")
def test_ray_cluster_comp_worker_node(
    mock_ray_cluster,
    mock_sleep,
):
    """Test the component functionality in the worker node"""
    mock_sleep.return_value = "1s"
    mock_ray_cluster.return_value = mock.MagicMock(workerpool_type="workerpool1")
    ray_comp_func(ray_task_cmd="test command")
    assert mock_ray_cluster.return_value.connect_to_node.called



import re
import os
import json
import time
import subprocess
from ray._private.test_utils import wait_for_condition
import logging


class RayCustomJobSpec:
    """
    Vertex AI populates an environment variable, CLUSTER_SPEC,
    on every replica to describe how the overall cluster is set up.
    This class retrieves some useful information from CLUSTER_SPEC,
    such as host name of the replica in workerpool0 and the open port.
    Explore more information in the following link:
    https://cloud.google.com/vertex-ai/docs/training/distributed-training#cluster-spec-format
    """

    def __init__(self):
        logging.basicConfig(level=logging.INFO)
        self._logger = logging.getLogger(__name__)

        try:
            self.cluster_config_str = os.environ.get("CLUSTER_SPEC")
            self.cluster_config_dict = json.loads(self.cluster_config_str)
            self.cluster_info = self.cluster_config_dict["cluster"]
            self.workerpool_type = self.cluster_config_dict["task"]["type"]
        except Exception as error:
            self._logger.error(f"{error}")
            raise RuntimeError(f"{error}")

    def get_head_node_name(self, workerpool_name: str = "workerpool0"):
        """Get the host name of the head node from the cluster spec

        Returns:
            str: cluster
        """
        try:
            host_name = self.cluster_info.get(workerpool_name, {})[0].split(":")[0]
        except Exception as error:
            raise ValueError(
                f"Fetching the host name of the head node from {workerpool_name} \
                    failed with error: {error}. Please check the cluster configuration \
                    {self.cluster_config_dict}"
            )
        return host_name

    def get_open_port(self):
        """Get the open port for nodes communication

        Args:
            cluster_config_dict (dict): Vertex custom job cluster configuration

        Returns:
            str: The port of the head node
        """
        if "open_ports" in self.cluster_config_dict:
            port = self.cluster_config_dict["open_ports"][0]
        else:
            # Use any port for the non-distributed job.
            port = 7777
        return port

    def get_head_node_info(self):
        open_port = self.get_open_port()
        head_node_name = self.get_head_node_name()
        head_node_address = f"{head_node_name}:{open_port}"
        return head_node_name, open_port, head_node_address


class RayCluster(RayCustomJobSpec):
    """
    This class establishes a Ray cluster in a Vertex custom training job cluster
    """

    def __init__(self):
        super().__init__()
        (
            self.head_node_name,
            self.open_port,
            self.head_node_address,
        ) = self.get_head_node_info()
        self._logger.info(f"The head node address is {self.head_node_address}")

    def run_cmd(self, cmd: str, logging: bool = True):
        """Run a bash command

        Args:
            cmd (str): Bash command to be executed.
            logging (bool): Whether to log the result or not. Defaults to True.

        Returns:
            str: The command result as a string
        """
        try:
            result = subprocess.run(
                cmd, shell=True, capture_output=True, text=True, check=True
            )
            self._logger.info(f"{cmd} executed successfully.")
            output = result.stdout
        except subprocess.CalledProcessError as error:
            raise RuntimeError(
                f"Command({cmd})failed with return code {error.returncode}: {error.stderr}"
            )
        if logging:
            self._logger.info(output)
        return output

    def start_ray_cluster(self):
        """Ran a Ray command to start a head node of a Ray cluster"""
        self.run_cmd(
            cmd=f"ray \
                      start \
                      --head \
                     --port={self.open_port} \
                     --redis-shard-ports={self.open_port}"
        )
        time.sleep(5)

    def check_nodes(
        self,
        node_address: str,
        target_num: int = 1,
    ):
        """Check if the number of nodes in the target address is over a certain amount

        Args:
            node_address(str): The node address for getting the node status.
            target_num (int, optional): The expected number of nodes.Default to 1.
        """
        cluster_status = self.run_cmd(
            cmd=f"ray \
                status \
                --address={node_address}",
            logging=False,
        )

        nodes_id = re.findall(r"1 node_.+", cluster_status)
        num_nodes = len(nodes_id)
        self._logger.debug(
            f"{num_nodes} of {target_num} found in the Ray cluster. They are {nodes_id}."
        )
        if num_nodes >= target_num:
            self._logger.info(f"{num_nodes} found in the Ray cluster {nodes_id}.")
            return True
        else:
            return False

    def wait_for_nodes(
        self,
        node_address: str,
        min_nodes: int = 1,
        timeout: int = 300,
        retry_interval_ms: int = 5000,
    ):
        """Wait util the number of nodes meets the requirement

        Args:
            node_address (str): The node address for getting the node status
            timeout (int, optional):  Maximum timeout in seconds.. Defaults to 300.
            retry_interval_ms (int, optional): Retry interval in milliseconds. Defaults to 5000.
        """
        self._logger.info("Waiting for nodes to be connected.")
        wait_for_condition(
            condition_predictor=self.check_nodes,
            timeout=timeout,
            retry_interval_ms=retry_interval_ms,
            **{"node_address": node_address, "target_num": min_nodes},
        )

    def connect_to_node(
        self, node_address: str, timeout: int = 300, retry_interval_ms: int = 5000
    ) -> None:
        """Connect worker node to the head node

        Args:
            node_address (str): The node address for getting the node status
            timeout (int, optional):  Maximum timeout in seconds.. Defaults to 300.
            retry_interval_ms (int, optional): Retry interval in milliseconds. Defaults to 5000.
        """
        wait_for_condition(
            condition_predictor=self.check_nodes,
            timeout=timeout,
            retry_interval_ms=retry_interval_ms,
            **{"node_address": node_address},
        )
        time.sleep(10)
        self.run_cmd(
            cmd=f"ray \
                     start \
                     --address={node_address}"
        )
        self._logger.info(f"{self.workerpool_type} has connected to {node_address}")

    def stop_process(self):
        """Stop Ray processes"""
        self.run_cmd(cmd="ray stop")

    def tear_down_cluster(self):
        """Tear down a Ray cluster."""
        self.run_cmd(cmd="ray down -y")



"""Unit tests for the KFP component that uploads a PyTorch model to Vertex AI"""
from unittest.mock import Mock, patch
import os
import pytest

from upload_pytorch_model.utils.gcs_helper import (
    split_bucket_object_file,
    stage_files_locally,
)

from upload_pytorch_model.utils.model_archiver import run_command

from upload_pytorch_model.component import upload_pytorch_model

upload_pytorch_model = upload_pytorch_model.python_func

# Fixtures and mocks


class MockStorageBlob:
    """Class that mocks a GCS Blob"""

    def __init__(self, name: str):
        self.name = name

    def download_to_filename(self, name):
        with open(name, "w") as f:
            f.write("")


class MockListBlobs(MockStorageBlob):
    """Class that mocks the Cloud Storage list_blob() method"""

    def __init__(self):
        self.list_blobs = []

    def add_blob(self, blob_to_add):
        self.list_blobs.append(blob_to_add)


def remove_file(filename):
    """Helper function to remove a file from local environment"""
    try:
        os.remove(filename)
    except FileNotFoundError:
        pass


@pytest.fixture
def mock_bucket_name():
    """Mocks a GCS bucket"""
    bucket_name = "test-model-bucket"
    return bucket_name


@pytest.fixture
def mock_object_name(tmp_path):
    """Mocks a GCS object"""
    return tmp_path


@pytest.fixture
def mock_model_name():
    """Mocks a PyTorch model"""
    model_name = "model_name.pt"
    yield model_name
    remove_file(model_name)


@pytest.fixture
def mock_pt_model(tmp_path, mock_model_name):
    """Mocks a PyTorch model and creates a mock GCS blob"""
    mock_model = Mock()
    mock_model.name = mock_model_name
    mock_model.path = str(tmp_path / mock_model.name)
    mock_model.blob = MockStorageBlob(name=mock_model.path)
    return mock_model


@pytest.fixture
def mock_mar_model(tmp_path):
    """Mocks a .mar file and GCS blob"""
    mock_mar = Mock()
    mock_mar.name = "model.mar"
    mock_mar.path = str(tmp_path / mock_mar.name)
    mock_mar.blob = MockStorageBlob(name=mock_mar.path)
    return mock_mar


@pytest.fixture
def mock_custom_handler(tmp_path):
    """Mocks a custom hanlder file and GCS blob"""
    mock_handler = Mock()
    mock_handler.name = "handler.py"
    mock_handler.path = str(tmp_path / mock_handler.name)
    mock_handler.blob = MockStorageBlob(name=mock_handler.path)
    yield mock_handler
    remove_file(mock_handler.name)


@pytest.fixture
def mock_requirements_file(tmp_path):
    """Mocks a requirements file and GCS blob"""
    mock_reqs = Mock()
    mock_reqs.name = "requirements.txt"
    mock_reqs.path = str(tmp_path / mock_reqs.name)
    mock_reqs.blob = MockStorageBlob(name=mock_reqs.path)
    yield mock_reqs
    remove_file(mock_reqs.name)


# Tests
def test_split_bucket_object_file(mock_bucket_name, mock_object_name, mock_model_name):
    """
    Tests that the split_bucket_object_file works
    when a file name is passed as input

    Args:
        mock_bucket_name: Mocks a GCS bucket
        mock_object_name: Mocks a GCS object
        mock_model_name: Mocks a file
    """
    bucket_name, object_name, file_name = split_bucket_object_file(
        path=f"gs://{mock_bucket_name}/{mock_object_name}/{mock_model_name}",
        file_name=mock_model_name,
    )

    assert bucket_name == mock_bucket_name
    assert object_name == str(mock_object_name)
    assert file_name == mock_model_name


def test_split_bucket_object_no_file(mock_bucket_name, mock_object_name):
    """
    Tests that the split_bucket_object_file works
    when no file name is passed as input

    Args:
        mock_bucket_name: Mocks a GCS bucket
        mock_object_name: Mocks a GCS object
    """
    bucket_name, object_name, file_name = split_bucket_object_file(
        path=f"gs://{mock_bucket_name}/{mock_object_name}/",
    )

    assert bucket_name == mock_bucket_name
    assert object_name == str(mock_object_name)
    assert file_name is None


def test_split_bucket_object_no_file_no_slash(mock_bucket_name, mock_object_name):
    """
    Tests that the split_bucket_object_filee works
    when no file name is passed as input and the object name
    is not properly parsed

    Args:
        mock_bucket_name: Mocks a GCS bucket
        mock_object_name: Mocks a GCS object
    """
    bucket_name, object_name, file_name = split_bucket_object_file(
        path=f"gs://{mock_bucket_name}/{mock_object_name}",
    )

    assert bucket_name == mock_bucket_name
    assert object_name == str(mock_object_name)
    assert file_name is None


def test_raise_error_model_name(mock_bucket_name, mock_object_name, mock_model_name):
    """
    Tests that the component raises a NameError if a wrong model name
    is used together with a pre-built container

    Args:
        mock_bucket_name: Mocks a GCS bucket
        mock_object_name: Mocks a GCS object
        mock_model_name: Mocks a file
    """
    with pytest.raises(NameError):
        upload_pytorch_model(
            project_id="test-project",
            location="europe-west2",
            gcs_model_path=f"gs://{mock_bucket_name}/{mock_object_name}/{mock_model_name}",
            model_name="test-name",
            model_display_name="test-name",
            local_model_file=mock_model_name,
            vertex_model=None,
            handler="text_classification",
        )


def test_raise_error_requirements(
    mock_bucket_name, mock_object_name, mock_model_name, mock_requirements_file
):
    """
    Tests that the component raises a TypeError if the model requires
    extra packages but the serving container used is the pre-built
    container, which does not allow for this

    Args:
        mock_bucket_name: Mocks a GCS bucket
        mock_object_name: Mocks a GCS object
        mock_model_name: Mocks a file
        mock_requirements_file: Mocks the requirements file
    """

    with pytest.raises(TypeError):
        upload_pytorch_model(
            project_id="test-project",
            location="europe-west2",
            gcs_model_path=f"gs://{mock_bucket_name}/{mock_object_name}/{mock_model_name}",
            model_display_name="test-name",
            local_model_file=mock_model_name,
            vertex_model=None,
            requirements=mock_requirements_file.path,
        )


@patch("subprocess.run")
def test_cmd_with_defaults(mock_subprocess, mock_model_name):
    """
    Tests that the run_command method correctly calls the torch-model-archiver
    when passed only default values

    Args:
        mock_subprocess: Mocks the subprocess.run execution
        mock_model_name: Mocks a model
    """
    # Prepare command for execution
    model_name = "model"
    handler = "default_handler"

    # Execute command
    cmd = run_command(
        model_name=model_name, serialized_file=mock_model_name, handler=handler
    )

    # Assert
    expected_cmd = [
        "torch-model-archiver",
        "-f",
        f"--model-name={model_name}",
        f"--serialized-file={mock_model_name}",
        f"--handler={handler}",
    ]

    assert cmd == expected_cmd


@patch("subprocess.run")
def test_cmd_with_extras(
    mock_subprocess, mock_model_name, mock_custom_handler, mock_requirements_file
):
    """
    Tests that the run_command method correctly calls the torch-model-archiver
    when additional optional arguments are passed

    Args:
        mock_subprocess: Mocks the subprocess.run execution
        mock_model_name: Mocks a model
        mock_custom_handler: Mocks the custom hanler file
        mock_requirements_file: Mocks the requirements file
    """
    # Prepare command for execution
    model_name = "model"
    extra_files = "file_1,file_2,path/file_3"
    version = "1.0"
    handler = mock_custom_handler.path
    requirements = mock_requirements_file.path

    # Execute command
    cmd = run_command(
        model_name=model_name,
        serialized_file=mock_model_name,
        extra_files=extra_files,
        version=version,
        handler=handler,
        requirements=requirements,
    )

    # Assert
    expected_cmd = [
        "torch-model-archiver",
        "-f",
        f"--model-name={model_name}",
        f"--serialized-file={mock_model_name}",
        f"--handler={handler}",
        f"--extra-files={extra_files}",
        f"--version={version}",
        f"--requirements-file={requirements}",
    ]

    assert cmd == expected_cmd


@patch("google.cloud.storage.Client")
def test_stage_one_file(mock_client, mock_bucket_name, mock_object_name, mock_pt_model):
    """
    Test that the stage_files_locally method can stage a single file
    when requested

    Args:
        mock_client: Mocks the GCS client google.cloud.storage.Client()
        mock_bucket_name: Mocks a GCS bucket
        mock_object_name: Mocks a GCS object
        mock_pt_model: Mocks a PyTorch model and blob
    """
    # Mock a GCS blob
    list_blobs = MockListBlobs()
    list_blobs.add_blob(mock_pt_model.blob)
    mock_client().list_blobs.return_value = list_blobs.list_blobs

    # Execute command to test
    stage_files_locally(
        bucket_name=mock_bucket_name,
        object_name=mock_object_name,
        file_name=mock_pt_model.name,
    )

    # Assert
    assert os.path.exists(mock_pt_model.name)


@patch("google.cloud.storage.Client")
def test_stage_all_files(
    mock_client,
    mock_bucket_name,
    mock_object_name,
    mock_pt_model,
    mock_custom_handler,
    mock_requirements_file,
):
    """
    Test that the stage_files_locally method can stage all files
    when requested

    Args:
        mock_client: Mocks the GCS client google.cloud.storage.Client()
        mock_bucket_name: Mocks a GCS bucket
        mock_object_name: Mocks a GCS object
        mock_pt_model: Mocks a PyTorch model and blob
        mock_custom_handler: Mocks a custom hanlder and blob
        mock_requirements_file: Mocks a requirements file and blob
    """
    # Mock multiple GCS blobs
    list_blobs = MockListBlobs()
    list_blobs.add_blob(mock_pt_model.blob)
    list_blobs.add_blob(mock_custom_handler.blob)
    list_blobs.add_blob(mock_requirements_file.blob)
    mock_client().list_blobs.return_value = list_blobs.list_blobs

    # Execute command to test
    extra_files = stage_files_locally(
        bucket_name=mock_bucket_name,
        object_name=mock_object_name,
        file_name="*",
        model_extra_files=True,
    )

    # Assert
    assert os.path.exists(mock_pt_model.name)
    assert os.path.exists(mock_custom_handler.name)
    assert os.path.exists(mock_requirements_file.name)
    assert (
        extra_files
        == f"{mock_pt_model.name},{mock_custom_handler.name},{mock_requirements_file.name}"
    )


@patch("google.cloud.storage.Client")
def test_stage_all_files_ignore(
    mock_client,
    mock_bucket_name,
    mock_object_name,
    mock_pt_model,
    mock_custom_handler,
    mock_requirements_file,
    mock_mar_model,
):
    """
    Test that the stage_files_locally method can stage all files
    apart from the ones to ignore when requested

    Args:
        mock_client: Mocks the GCS client google.cloud.storage.Client()
        mock_bucket_name: Mocks a GCS bucket
        mock_object_name: Mocks a GCS object
        mock_pt_model: Mocks a PyTorch model and blob
        mock_custom_handler: Mocks a custom hanlder and blob
        mock_requirements_file: Mocks a requirements file and blob
        mock_mar_model: Mocks a .mar file and blob
    """
    # Mock multiple GCS blobs
    list_blobs = MockListBlobs()
    list_blobs.add_blob(mock_pt_model.blob)
    list_blobs.add_blob(mock_custom_handler.blob)
    list_blobs.add_blob(mock_requirements_file.blob)
    list_blobs.add_blob(mock_mar_model.blob)  # This file will be ignored
    mock_client().list_blobs.return_value = list_blobs.list_blobs

    # Execute command to test
    extra_files = stage_files_locally(
        bucket_name=mock_bucket_name,
        object_name=mock_object_name,
        file_name="*",
        ignore_files=[mock_mar_model.name],
        model_extra_files=True,
    )

    # Assert
    assert os.path.exists(mock_pt_model.name)
    assert os.path.exists(mock_custom_handler.name)
    assert os.path.exists(mock_requirements_file.name)
    assert os.path.exists(mock_mar_model.name) is False  # This file should not exist
    assert (
        extra_files
        == f"{mock_pt_model.name},{mock_custom_handler.name},{mock_requirements_file.name}"
    )



import subprocess
from upload_pytorch_model.utils.gcs_helper import (
    stage_files_locally,
    split_bucket_object_file,
    upload_to_gcs,
)


def run_command(
    model_name: str,
    serialized_file: str,
    handler: str,
    version: str = None,
    extra_files: str = None,
    requirements: str = None,
):
    """Execute command to create torch model archiver"""

    cmd = [
        "torch-model-archiver",
        "-f",
        f"--model-name={model_name}",
        f"--serialized-file={serialized_file}",
        f"--handler={handler}",
    ]

    # TODO: There has to be a neater way of doing this
    if extra_files and len(extra_files) > 0:
        cmd += [f"--extra-files={extra_files}"]
    if version:
        cmd += [f"--version={version}"]
    if requirements:
        cmd += [f"--requirements-file={requirements}"]

    process = subprocess.run(cmd, capture_output=True, shell=False)
    process.check_returncode()

    return cmd


def create_model_archiver(
    gcs_model_path: str,
    model_name: str,
    version: str,
    local_model_file: str,
    handler: str,
    extra_files_staging_path: str = None,
    extra_files_to_ignore: list = [],
    extra_prediction_files: str = None,
    requirements: str = None,
    custom_tokenizer_path: str = None,
    custom_tokenizer_name: str = None,
):
    """
    Create model archiver file. For this, it needs to stage locally the following files:

        1. Model artifact + model files in same folder (outputted by default by HuggingFace)
        2. Optionally, it can download:
            2.1 Custom handler file
            2.2 Requirements file
            2.3 Custom tokemizer artifact
            2.4 Additional files required for prediction
    """
    model_bucket_name, model_object_name, _ = split_bucket_object_file(
        path=gcs_model_path
    )
    if extra_files_staging_path:
        (
            extra_files_staging_bucket,
            extra_files_staging_object,
            _,
        ) = split_bucket_object_file(path=extra_files_staging_path)

    print("Staging model file locally...")
    stage_files_locally(
        bucket_name=model_bucket_name,
        object_name=model_object_name,
        file_name=local_model_file,
    )

    print("Staging additional model files locally...")
    extra_files = stage_files_locally(
        bucket_name=model_bucket_name,
        object_name=model_object_name,
        ignore_files=extra_files_to_ignore + [local_model_file],
        model_extra_files=True,
    )

    if handler.endswith(".py"):  # Custom handler script
        print("Staging custom handler file locally")
        stage_files_locally(
            bucket_name=extra_files_staging_bucket,
            object_name=extra_files_staging_object,
            file_name=handler,
        )
    if requirements:
        print("Staging model requirements file locally")
        stage_files_locally(
            bucket_name=extra_files_staging_bucket,
            object_name=extra_files_staging_object,
            file_name=requirements,
        )

    if extra_prediction_files:
        extra_files = extra_files + "," + extra_prediction_files
        print("Staging extra prediction files locally")
        for prediction_file in extra_prediction_files.split(","):
            stage_files_locally(
                bucket_name=extra_files_staging_bucket,
                object_name=extra_files_staging_object,
                file_name=prediction_file,
            )

    if custom_tokenizer_path:
        print("Staging custom tokenizer locally")
        (
            custom_tokenizer_bucket,
            custom_tokenizer_object,
            custom_tokenizer_name,
        ) = split_bucket_object_file(
            path=extra_files_staging_path, file_name=custom_tokenizer_name
        )
        stage_files_locally(
            bucket_name=custom_tokenizer_bucket,
            object_name=custom_tokenizer_object,
            file_name=custom_tokenizer_name,
        )
        extra_files = extra_files + "," + custom_tokenizer_name

    if len(extra_files) > 0 and extra_files[0] == ",":
        extra_files = extra_files[1:]

    print("Run TAR packaging")
    cmd = run_command(
        model_name=model_name,
        version=version,
        serialized_file=local_model_file,
        handler=handler,
        extra_files=extra_files,
        requirements=requirements,
    )

    print("Uploading to GCS")
    upload_to_gcs(
        bucket_name=model_bucket_name,
        object_name=model_object_name,
        model_name=model_name,
    )



import os
from pathlib import Path
from kfp.dsl import component

_COMPONENTS_CONTAINER_REGISTRY = os.environ.get(
    "COMPONENTS_CONTAINER_REGISTRY",
    "europe-west1-docker.pkg.dev/dt-playgrounds-mlops/turbo-templates-containers",
)
_IMAGE_NAME = str(Path(__file__).parents[1].name)
_COMPONENTS_IMAGE_TAG = os.environ.get("COMPONENTS_IMAGE_TAG", "latest")


@component(
    base_image="python:3.9-slim-buster",
    target_image=f"{_COMPONENTS_CONTAINER_REGISTRY}/{_IMAGE_NAME}:{_COMPONENTS_IMAGE_TAG}",  # noqa: E501
    output_component_file=os.path.splitext(__file__)[0] + ".yaml",
)
def ray_cluster(
    ray_task_cmd: str,
    node_config_cmd: str = None,
    min_nodes: int = 1,
    cluster_op_timeout: int = 300,
    retry_interval_ms: int = 5000,
    nodes_alive_time: int = 24 * 60 * 60,
):
    """Deploy a Ray cluster and run distributed tasks

    Args:
        ray_task_cmd (str): The command to run the Ray task(s).
        node_config_cmd (str, optional): Command for additional setup that will be executed on all nodes in the cluster.
            For example, installing the dependencies or cloning a codebase. Defaults to None.
        min_nodes (int, optional): The minimum nodes required in the Ray cluster. Defaults to 1.
        cluster_op_timeout (int, optional):  Maximum timeout in seconds of any Ray cluster establish jobs.
            Defaults to 300.
        retry_interval_ms (int, optional): Retry interval in milliseconds of any Ray cluster establish jobs.
            Defaults to 5000.
        nodes_alive_time (int, optional):The amount of time that a node is alive. Defaults to 24*60*60 (1 day).
    """
    import time
    from ray.job_submission import JobSubmissionClient
    from ray_cluster.utils.utils import RayCluster

    ray_cluster = RayCluster()
    if node_config_cmd is not None:
        ray_cluster.run_cmd(cmd=node_config_cmd)

    if ray_cluster.workerpool_type == "workerpool0":
        # Initialising the ray cluster
        ray_cluster.start_ray_cluster()

        ray_cluster.wait_for_nodes(
            node_address=ray_cluster.head_node_address,
            min_nodes=min_nodes,
            timeout=cluster_op_timeout,
            retry_interval_ms=retry_interval_ms,
        )
        time.sleep(5)
        ray_cluster._logger.info(
            "A Ray cluster has been created. Running the custom job."
        )

        # Run ray tasks
        ray_job_client = JobSubmissionClient(ray_cluster.head_node_address)
        submission_id = ray_job_client.submit_job(
            entrypoint=ray_task_cmd,
        )
        ray_cluster._logger.info(
            "Job has been submitted successfully!\n Here are the logs:"
        )
        ray_cluster.run_cmd(
            cmd=f"ray \
                            job \
                            logs \
                            {submission_id} \
                            --follow"
        )
        ray_job_info = ray_job_client.get_job_info(submission_id)
        job_status = ray_job_info.status
        if job_status == "FAILED":
            raise RuntimeError("Job failed")
        else:
            ray_cluster._logger.info(
                f"Job {job_status} with the following information:"
            )
            ray_cluster._logger.info(ray_job_info)

    else:
        # Connect worker nodes to the head node
        ray_cluster.connect_to_node(
            node_address=ray_cluster.head_node_address,
            timeout=cluster_op_timeout,
            retry_interval_ms=retry_interval_ms,
        )
        time.sleep(nodes_alive_time)



import os
from google.cloud import storage


def stage_files_locally(
    bucket_name: str,
    object_name: str,
    file_name: str = "*",
    ignore_files: list = [],
    model_extra_files: bool = False,
):
    """Stage prediction files locally"""

    if model_extra_files:
        extra_files = ""

    storage_client = storage.Client()
    blobs = storage_client.list_blobs(bucket_name, prefix=object_name)
    for blob in blobs:
        blob_name = blob.name.split(f"{object_name}/")[1]
        if blob_name in ignore_files or blob_name == "":
            continue
        if blob_name.endswith("/"):
            os.makedirs(blob_name[:-1], exist_ok=True)
        elif blob_name == file_name or file_name == "*":
            blob.download_to_filename(blob_name)
            print(f"Downloaded {blob_name}")
            if model_extra_files:
                extra_files += f"{blob_name}" + ","

    if model_extra_files:
        extra_files = extra_files.rstrip(",")
        return extra_files


def upload_to_gcs(bucket_name: str, object_name: str, model_name: str):
    """Upload packaged .mar file to GCS"""

    if object_name.endswith("/"):
        object_name = object_name[:-1]

    storage_client = storage.Client()
    bucket = storage_client.bucket(bucket_name)
    destination_blob_name = f"{object_name}/{model_name}.mar"
    blob = bucket.blob(destination_blob_name)
    blob.upload_from_filename(f"{model_name}.mar")

    return destination_blob_name


def split_bucket_object_file(path: str, file_name: str = None) -> (str, str, str):
    """Splits a GCS path and file name into bucket, object and file"""
    bucket_name = path.split("gs://")[1].split("/")[0]
    full_object_name = path.split(f"gs://{bucket_name}/")[1]
    if file_name:
        object_name = full_object_name.split(file_name)[0]
    else:
        object_name = full_object_name
    if object_name.endswith("/"):
        object_name = object_name[:-1]
    return bucket_name, object_name, file_name



from unittest.mock import Mock

import pytest
import pandas as pd
import numpy as np

from component import (
    gpu_accelerated_shap_for_xgboost as gpu_accelerated_shap_for_xgboost_component,
)
from component import calculate_shap_values


gpu_accelerated_shap_for_xgboost = (
    gpu_accelerated_shap_for_xgboost_component.python_func
)


@pytest.fixture
def mock_input_data(tmp_path, monkeypatch):
    # Mock component method inputs
    mock_input_data = Mock()
    mock_input_data.path = str(tmp_path / "input.csv")
    input_df = pd.DataFrame([{"a": 1, "b": 2, "c": 3}])
    input_df.to_csv(mock_input_data.path, index=False)
    return mock_input_data


@pytest.fixture
def mock_joblib(monkeypatch):
    # Mock model load from gcs
    mock_joblib = Mock()
    mock_joblib.load.return_value = Mock()
    monkeypatch.setattr("utils.joblib", mock_joblib)
    return mock_joblib


@pytest.fixture
def mock_calculate_shap(monkeypatch):
    # Mock shap calculation
    mock_shap_values = np.random.rand(1, 3)
    mock_calculate_shap = Mock()
    mock_calculate_shap.return_value = mock_shap_values
    monkeypatch.setattr("component.calculate_shap_values", mock_calculate_shap)
    return mock_calculate_shap


@pytest.fixture
def mock_model(tmp_path):
    mock_model = Mock()
    mock_model.path = str(tmp_path / "model.joblib")
    return mock_model


@pytest.fixture
def mock_output_data(tmp_path):
    mock_output_data = Mock()
    mock_output_data.path = str(tmp_path / "output")
    return mock_output_data


def test_gpu_accelerated_shap_for_xgboost(
    mock_input_data, mock_joblib, mock_model, mock_calculate_shap, mock_output_data
):
    """
    Tests the following flow from the component:
        1. Model is loaded in
        2. Input data is read in a dataframe
        3. Shap values are retrieved
        4. Shap values are converted into a dataframe and saved to a file
    """
    # Arrange
    # Create expected output df
    input_df = pd.read_csv(mock_input_data.path)
    mock_shap_values = mock_calculate_shap()
    expected_output_df = pd.DataFrame(mock_shap_values, columns=list(input_df.columns))

    # Act
    gpu_accelerated_shap_for_xgboost(mock_input_data, mock_model, mock_output_data)

    # Assert
    output_df = pd.read_csv(mock_output_data.path)
    pd.util.testing.assert_frame_equal(
        output_df.reset_index(drop=True), expected_output_df.reset_index(drop=True)
    )


def test_calculate_shap_values_gpu_enabled(monkeypatch, mock_input_data):
    """Tests whether shap explainer is instantiated with GPU config"""
    # Arrange
    mock_xgb_model = Mock()
    input_data = pd.read_csv(mock_input_data.path)

    mock_shap = Mock()
    monkeypatch.setattr("component.shap", mock_shap)

    # Act
    calculate_shap_values(mock_xgb_model, input_data, enable_gpu=True)

    # Assert
    mock_xgb_model.get_booster.assert_called_once()


def test_calculate_shap_values_gpu_disbaled(monkeypatch, mock_input_data):
    """Tests whether shap explainer is instantiated without GPU (only CPU) config"""
    # Arrange
    mock_xgb_model = Mock()
    input_data = pd.read_csv(mock_input_data.path)

    mock_shap = Mock()
    monkeypatch.setattr("component.shap", mock_shap)

    # Act
    calculate_shap_values(mock_xgb_model, input_data, enable_gpu=False)

    # Assert
    mock_xgb_model.get_booster.assert_not_called()



from pathlib import Path
from kfp.components import load_component_from_file

# Get the current directory as a Path object
current_dir = Path(__file__).resolve().parent

# Find all directories in the current directory that are not prefixed with "_"
components = [
    x for x in current_dir.iterdir() if x.is_dir() and not x.name.startswith("_")
]

# Load the components and create global variables for each of them
for component in components:
    component_yaml_path = component / f"{component.name}/component.yaml"
    globals()[component.name] = load_component_from_file(component_yaml_path)

# Define __all__ as the list of loaded components
__all__ = [component.name for component in components]



import time
import logging
from typing import Union
import shap
import joblib
import pandas as pd
import xgboost as xgb


def load_joblib_model(model_path: str):
    return joblib.load(model_path)


def calculate_shap_values(
    model: Union[xgb.XGBClassifier, xgb.XGBRegressor],
    X: pd.DataFrame,
    enable_gpu: bool,
):
    start = time.process_time()
    if enable_gpu:
        logging.info(
            f"Using GPU to calculate shap values on dataframe with shape {X.shape} "
        )
        model_gpu = model.get_booster()
        model_gpu.set_param({"predictor": "gpu_predictor"})
        explainer = shap.TreeExplainer(model=model_gpu)
    else:
        logging.info(
            f"Using CPU to calculate shap values on dataframe with shape {X.shape} "
        )
        explainer = shap.TreeExplainer(model=model)
    shap_values = explainer.shap_values(X)
    time_taken = time.process_time() - start
    logging.info(f"Took {time_taken} to calculate shap values")
    return shap_values



"""KFP component to load PyTorch model from HuggingFace into GCS"""
import os
from pathlib import Path
from kfp.dsl import Output, Artifact, component

_COMPONENTS_CONTAINER_REGISTRY = os.environ.get(
    "COMPONENTS_CONTAINER_REGISTRY",
    "europe-west2-docker.pkg.dev/dt-pc-dev/dt-pc-dev-components-repository",
)
_IMAGE_NAME = str(Path(__file__).parents[1].name)
_COMPONENTS_IMAGE_TAG = os.environ.get("COMPONENTS_IMAGE_TAG", "latest")


@component(
    base_image="python:3.9-slim-buster",
    target_image=f"{_COMPONENTS_CONTAINER_REGISTRY}/{_IMAGE_NAME}:{_COMPONENTS_IMAGE_TAG}",  # noqa: E501
    output_component_file=os.path.splitext(__file__)[0] + ".yaml",
)
def load_hf_model(
    model_class_name: str,
    tokenizer_class_name: str,
    config_name: str,
    model_name: str,
    hf_gcs_path: Output[Artifact],
    model_args: dict = {},
    kwargs: dict = {},
):
    """
    Loads PyTorch model, config and tokenizer from HuggingFace
    and stores artefacts in GCS

    Args:
        model_class_name [str]: HuggingFace class to import model from (ie: AutoModelForCasualLM)
        tokenizer_class_name [str]: HuggingFace class to import tokenizer from (ie: AutoTokenizer)
        config_name [str]: HuggingFace class to import config from (ie: AutoConfig)
        model_name [str]: HuggingFace repo model name (ie: bert-base-uncased)
        model_args [dict]: Optinal additional model arguments passed to model class
        kwargs [dict]: Optional additional keywords required by model, config and tokenizer classes.
            The required format is:
                kwargs = {
                    "model":
                        {
                            "key_model_arg_1":value_1,
                            "key_model_arg_2":value_2,
                        },
                    "config":
                        {
                            "key_config_arg_1":value_1,
                            "key_config_arg_2":value_2,
                        },
                    "token":
                        {
                            "key_token_arg_1":value_1,
                            "key_token_arg_2":value_2,
                        },
                }

    Returns:
        hf_gcs_path [Artifact]: Path to GCS objects with model artifacts
    """
    # Import transformers and required classes
    import transformers

    # Raise error in TensorFlow model selected
    if model_class_name.startswith("TF"):
        raise NameError(
            """
        This component only accepts PyTorch models but you selected
        a TensorFlow model. Please use the other component in this library for TensorFlow models
        """
        )

    model_class = getattr(transformers, model_class_name)
    config_class = getattr(transformers, config_name)
    tokenizer_class = getattr(transformers, tokenizer_class_name)

    # Load optional additional args
    config_kwargs = kwargs.get("config", {})
    model_kwargs = kwargs.get("model", {})
    token_kwargs = kwargs.get("token", {})

    # Load model config
    model_config = config_class.from_pretrained(model_name, **config_kwargs)

    # Load model
    model = model_class.from_pretrained(
        model_name,
        config=model_config,
        *model_args,
        **model_kwargs,
    )

    # Load tokenizer
    tokenizer = tokenizer_class.from_pretrained(model_name, **token_kwargs)

    # Save model artifacts in GCS
    model.save_pretrained(save_directory=hf_gcs_path.path)
    tokenizer.save_pretrained(save_directory=hf_gcs_path.path)



"""Unit tests for component that loads PyTorch models from HuggingFace"""
import os
import pytest
from unittest.mock import Mock

from tempfile import TemporaryDirectory

from component import (
    load_hf_model as load_hf_model_component,
)

load_hf_model = load_hf_model_component.python_func


def test_loads_model_config():
    """Test model config is correctly loaded"""
    with TemporaryDirectory() as tmp_dir:
        hf_gcs_path = Mock(path=tmp_dir)
        load_hf_model(
            "AutoModel", "AutoTokenizer", "AutoConfig", "bert-base-uncased", hf_gcs_path
        )
        assert os.path.exists(os.path.join(tmp_dir, "config.json"))


def test_loads_model():
    """Test model is correctly loaded"""
    with TemporaryDirectory() as tmp_dir:
        hf_gcs_path = Mock(path=tmp_dir)
        load_hf_model(
            "AutoModel", "AutoTokenizer", "AutoConfig", "bert-base-uncased", hf_gcs_path
        )
        assert os.path.exists(os.path.join(tmp_dir, "pytorch_model.bin"))


def test_loads_tokenizer():
    """Test tokenizer is correctly loaded"""
    with TemporaryDirectory() as tmp_dir:
        hf_gcs_path = Mock(path=tmp_dir)
        load_hf_model(
            "AutoModel", "AutoTokenizer", "AutoConfig", "bert-base-uncased", hf_gcs_path
        )
        assert os.path.exists(os.path.join(tmp_dir, "tokenizer.json"))


def test_loads_additional_args():
    """Test additional arguments are correctly parsed"""
    with TemporaryDirectory() as tmp_dir:
        hf_gcs_path = Mock(path=tmp_dir)
        model_args = {"output_attentions": True}
        kwargs = {
            "model": {"force_download": False},
            "config": {"max_length": 128},
            "token": {"do_lower_case": True},
        }
        load_hf_model(
            "AutoModel",
            "AutoTokenizer",
            "AutoConfig",
            "bert-base-uncased",
            hf_gcs_path,
            model_args=model_args,
            kwargs=kwargs,
        )
        assert os.path.exists(os.path.join(tmp_dir, "config.json"))
        assert os.path.exists(os.path.join(tmp_dir, "pytorch_model.bin"))
        assert os.path.exists(os.path.join(tmp_dir, "tokenizer.json"))


def test_tensorflow_raises_error():
    """Test TensorFlow models correctly raise an error"""
    with TemporaryDirectory() as tmp_dir:
        hf_gcs_path = Mock(path=tmp_dir)
        with pytest.raises(NameError):
            load_hf_model(
                "TFAutoModel",
                "AutoTokenizer",
                "AutoConfig",
                "bert-base-uncased",
                hf_gcs_path,
            )



import logging
import os
from pathlib import Path

import numpy as np
import tiktoken
from kfp.dsl import Artifact, Dataset, Input, Output, component

_COMPONENTS_CONTAINER_REGISTRY = os.environ.get(
    "COMPONENTS_CONTAINER_REGISTRY",
    "europe-west2-docker.pkg.dev/dt-pc-dev/dt-pc-dev-components-repository",
)
_IMAGE_NAME = str(Path(__file__).parents[1].name)
_COMPONENTS_IMAGE_TAG = os.environ.get("COMPONENTS_IMAGE_TAG", "latest")


@component(
    base_image="python:3.9-slim",
    packages_to_install=[
        "numpy==1.21.6",
        "google-cloud-storage==1.44.0",
        "tiktoken==0.3.0",
    ],
    target_image=f"{_COMPONENTS_CONTAINER_REGISTRY}/{_IMAGE_NAME}:{_COMPONENTS_IMAGE_TAG}",
    output_component_file=str(Path(__file__).with_suffix(".yaml")),
)
def gpt_tokenize(
    input_data: Input[Artifact],
    train_dataset: Output[Dataset],
    val_dataset: Output[Dataset],
    train_size: float = 0.9,
    encoding_type: str = "gpt2",
):
    """
    Reads the scraped data and generates a tokenized training and validation dataset

    Args:
        input_data (Input[Artifact]): Input text file contain training data
        train_dataset (Output[Dataset]): Training ids as a dataset object
        val_dataset (Output[Dataset]): Validation ids as a dataset object
        train_size (float): A value between 0.0 and 1.0 to represent the proportion of the dataset
            to include in the train split
        encoding_type (str): Encodings used by OpenAI models. Defaults to "gpt2", alternatives include
            "cl100k_base" or "p50k_base". See
            https://github.com/openai/openai-cookbook/blob/main/examples/How_to_count_tokens_with_tiktoken.ipynb
            for more details.
    """

    with open(
        os.path.join(input_data.path, "input.txt"), "r", encoding="utf-8"
    ) as file:
        data = file.read()

    logging.info(
        "Splitting data. Training set will be %s%% of the full data", train_size * 100
    )
    n = len(data)
    train_data = data[: int(n * train_size)]
    val_data = data[int(n * train_size) :]

    logging.info("Using %s encodings to encode data", encoding_type)
    enc = tiktoken.get_encoding(encoding_type)
    train_ids = enc.encode_ordinary(train_data)
    val_ids = enc.encode_ordinary(val_data)
    logging.info("train has %s tokens", len(train_ids))
    logging.info("val has %s tokens", len(val_ids))

    # export to bin files
    train_ids = np.array(train_ids, dtype=np.uint16)
    val_ids = np.array(val_ids, dtype=np.uint16)

    os.makedirs(train_dataset.path, exist_ok=True)
    os.makedirs(val_dataset.path, exist_ok=True)

    train_dataset_uri = os.path.join(train_dataset.path, "train.bin")
    val_dataset_uri = os.path.join(val_dataset.path, "val.bin")
    train_ids.tofile(train_dataset_uri)
    logging.info("Saving train_ids as %s ", train_dataset_uri)
    val_ids.tofile(val_dataset_uri)
    logging.info("Saving test_ids as %s ", val_dataset_uri)



import os
from pathlib import Path
from kfp.dsl import component, Model, Input, Output, Artifact

_COMPONENTS_CONTAINER_REGISTRY = os.environ.get(
    "COMPONENTS_CONTAINER_REGISTRY",
    "europe-west2-docker.pkg.dev/dt-pc-dev/dt-pc-dev-components-repository",
)
_IMAGE_NAME = str(Path(__file__).parent.name)
_COMPONENTS_IMAGE_TAG = os.environ.get("COMPONENTS_IMAGE_TAG", "latest")


@component(
    base_image="nvidia/cuda:11.3.0-base-ubuntu18.04",
    target_image=f"{_COMPONENTS_CONTAINER_REGISTRY}/{_IMAGE_NAME}:{_COMPONENTS_IMAGE_TAG}",  # noqa: E501
    output_component_file=os.path.splitext(__file__)[0] + ".yaml",
)
def gpu_accelerated_shap_for_xgboost(
    input_data: Input[Artifact],
    model: Input[Model],
    output_shap_values: Output[Artifact],
    enable_gpu: bool = True,
):
    """
    This function is a kfp component that explains an XGBoost model's predictions
    using the SHAP library.
    This component uses a base_image that has nvidia/cuda drivers installed.

    Args:
        input_data (Input[Artifact]): CSV file with pre-processed data stored on GCS.
        model (Input[Model]): The XGBoost model.joblib file stored in GCS in the format
            "gs://my-bucket/path/to/model/model.joblib".
        output_feature_attributions (Output[Artifact]): CSV file with output
            feature attribution data stored on GCS.
        enable_gpu (bool): Boolean flag to enable computation using GPU
    """
    import logging
    import pandas as pd
    from xgboost_shap_gpu.utils import calculate_shap_values, load_joblib_model

    logging.getLogger().setLevel(logging.INFO)

    # Get model and data
    xgb_model = load_joblib_model(model_path=model.path)
    input_df = pd.read_csv(input_data.path)

    # Calculate shap values with gpu
    shap_values = calculate_shap_values(xgb_model, input_df, enable_gpu)

    # Create column names
    feature_names = list(input_df.columns)

    # Create final df
    shap_df = pd.DataFrame(shap_values, columns=feature_names)

    # Upload to GCS
    output_shap_values.path += ".csv"
    shap_df.to_csv(output_shap_values.path, index=False)
    logging.info(f"Successfully uploaded to {output_shap_values.path}")



import pytest
import os
import json


@pytest.fixture
def mock_cluster_spec():
    cluster_spec = {
        "cluster": {
            "workerpool0": ["machine-name-workerpool0-0:2222"],
            "workerpool1": [
                "machine-name-workerpool1-0:2222",
                "machine-name-workerpool1-1:2222",
            ],
        },
        "task": {"type": "workerpool0", "index": 0, "trial": "TRIAL_ID"},
        "open_ports": ["open_port"],
    }
    os.environ["CLUSTER_SPEC"] = json.dumps(cluster_spec)


@pytest.fixture()
def mock_empty_cluster_spec():
    os.environ["CLUSTER_SPEC"] = json.dumps({})


@pytest.fixture()
def mock_cluster_spec_without_cluster():
    cluster_spec = {
        "task": {"type": "workerpool0", "index": 0, "trial": "TRIAL_ID"},
        "open_ports": ["open_port"],
    }
    os.environ["CLUSTER_SPEC"] = json.dumps(cluster_spec)


@pytest.fixture()
def mock_cluster_spec_without_task():
    cluster_spec = {
        "cluster": {
            "workerpool0": ["machine-name-workerpool0-0:2222"],
            "workerpool1": [
                "machine-name-workerpool1-0:2222",
                "machine-name-workerpool1-1:2222",
            ],
        },
        "open_ports": ["open_port"],
    }
    os.environ["CLUSTER_SPEC"] = json.dumps(cluster_spec)


@pytest.fixture
def mock_cluster_spec_without_workerpool():
    cluster_spec = {
        "cluster": {},
        "task": {"type": "workerpool0", "index": 0, "trial": "TRIAL_ID"},
        "open_ports": ["open_port"],
    }
    os.environ["CLUSTER_SPEC"] = json.dumps(cluster_spec)


@pytest.fixture
def mock_cluster_spec_without_open_port():
    cluster_spec = {
        "cluster": {
            "workerpool0": ["machine-name-workerpool0-0:2222"],
            "workerpool1": [
                "machine-name-workerpool1-0:2222",
                "machine-name-workerpool1-1:2222",
            ],
        },
        "task": {"type": "workerpool0", "index": 0, "trial": "TRIAL_ID"},
    }
    os.environ["CLUSTER_SPEC"] = json.dumps(cluster_spec)



"""Test module for gpt_tokenize component"""
import os
from unittest.mock import MagicMock, Mock

import pytest
from gpt_tokenize import gpt_tokenize


@pytest.fixture
def mock_input_data(tmp_path):
    # Mock input_data Artifact
    mock_input_text = MagicMock()
    mock_input_text.path = str(tmp_path)
    input_data = "Lorem ipsum dolor sit amet, consectetur adipiscing elit, \
        sed do eiusmod tempor incididunt ut labore et dolore magna aliqua."
    with open(f"{mock_input_text.path}/input.txt", "w", encoding="utf-8") as file:
        file.write(input_data)
    return mock_input_text


@pytest.fixture
def mock_train_dataset(tmp_path):
    # Mock output train_dataset artifact
    mock_train_dataset = Mock()
    mock_train_dataset.path = str(tmp_path)
    return mock_train_dataset


@pytest.fixture
def mock_val_dataset(tmp_path):
    # Mock output val_dataset artifact
    mock_val_dataset = Mock()
    mock_val_dataset.path = str(tmp_path)
    return mock_val_dataset


def test_correct_train_test_files_created(
    mock_input_data, mock_train_dataset, mock_val_dataset, monkeypatch
):
    """
    Tests that train and val .bin files are created when running the component logic by:
        1. Loading in mocked input data
        2. Splitting the data into train and val
        3. Encoding the datasets
        4. Exporting them to bin files
    """
    # Arrange
    mock_tiktoken = MagicMock()
    mock_tiktoken.get_encoding.encode_ordinary.return_value = MagicMock()
    monkeypatch.setattr("gpt_tokenize.component.tiktoken", mock_tiktoken)

    # Act
    gpt_tokenize.python_func(mock_input_data, mock_train_dataset, mock_val_dataset)

    # Assert
    assert os.path.exists(mock_train_dataset.path + "/train.bin")
    assert os.path.exists(mock_val_dataset.path + "/val.bin")



"""Unit tests for all utils functions"""
import pytest
import subprocess
from unittest import mock
from ray_cluster.utils.utils import RayCustomJobSpec, RayCluster


@pytest.mark.parametrize(
    "cluster_spec, result",
    [
        (pytest.lazy_fixture("mock_cluster_spec_without_cluster"), "error"),
        (pytest.lazy_fixture("mock_cluster_spec_without_task"), "error"),
        (pytest.lazy_fixture("mock_cluster_spec_without_cluster"), "error"),
    ],
)
def test_ray_custom_job_spec_init_failure(cluster_spec, result: str):
    """Testing if initialising RayCustomJobSpec class failed as expected

    Args:
        result (str): The expected result.
    """
    if result == "error":
        with pytest.raises(RuntimeError):
            custom_job_spec = RayCustomJobSpec()


def test_get_head_node_name(mock_cluster_spec):
    """Test get_head_node_name() in RayCustomJobSpec"""
    custom_job_spec = RayCustomJobSpec()
    assert custom_job_spec.get_head_node_name() == "machine-name-workerpool0-0"


def test_get_head_node_name_failure(mock_cluster_spec_without_workerpool):
    """Test if get_head_node_name() in RayCustomJobSpec failed as expected"""
    with pytest.raises(ValueError):
        custom_job_spec = RayCustomJobSpec()
        custom_job_spec.get_head_node_name()


def test_get_open_port_exist(mock_cluster_spec):
    """Test if get_open_port() in RayCustomJobSpec returns the value of the open_port filed"""
    custom_job_spec = RayCustomJobSpec()
    assert custom_job_spec.get_open_port() == "open_port"


def test_get_open_port_not_exist(mock_cluster_spec_without_open_port):
    """Test get_open_port() in RayCustomJobSpec returns the default value(7777)"""
    custom_job_spec = RayCustomJobSpec()
    assert custom_job_spec.get_open_port() == 7777


def test_get_head_node_info(mock_cluster_spec):
    """Test get_head_node_info() in RayCustomJobSpec"""
    custom_job_spec = RayCustomJobSpec()
    assert custom_job_spec.get_head_node_info() == (
        "machine-name-workerpool0-0",
        "open_port",
        "machine-name-workerpool0-0:open_port",
    )


@mock.patch("subprocess.run")
def test_run_command(mock_run, mock_cluster_spec):
    """Test run_command() in RayCluster"""
    cmd = "echo 'A test command.'"
    mock_run.return_value = mock.MagicMock(stdout="A test command.")
    ray_cluster = RayCluster()
    assert ray_cluster.run_cmd(cmd=cmd) == "A test command."


@mock.patch("subprocess.run")
def test_run_command_failure(mock_run, mock_cluster_spec):
    """Test if run_command() in RayCluster failed as expected"""
    cmd = "A wrong command"
    mock_run.side_effect = subprocess.CalledProcessError(1, cmd[0])
    ray_cluster = RayCluster()
    with pytest.raises(RuntimeError):
        ray_cluster.run_cmd(cmd=cmd)


@mock.patch("subprocess.run")
def test_check_nodes(mock_run, mock_cluster_spec):
    """Test check_nodes() in RayCluster"""
    mock_run.return_value = mock.MagicMock(stdout="1 node_name")
    ray_cluster = RayCluster()
    assert ray_cluster.check_nodes(node_address="mock_address")


@mock.patch("subprocess.run")
def test_check_nodes_failure(mock_run, mock_cluster_spec):
    """Test if check_nodes() in RayCluster failed as expected"""
    mock_run.return_value = mock.MagicMock(stdout="")
    ray_cluster = RayCluster()
    assert not ray_cluster.check_nodes(node_address="mock_address")


@mock.patch("ray_cluster.utils.utils.wait_for_condition")
def test_wait_for_nodes(mock_wait_for_condition, mock_cluster_spec):
    """Test wait_for_nodes() in RayCluster"""
    ray_cluster = RayCluster()
    mock_wait_for_condition.return_value = "done"
    ray_cluster.wait_for_nodes(node_address="mock_address")
    assert mock_wait_for_condition.called


@mock.patch("time.sleep")
@mock.patch("subprocess.run")
@mock.patch("ray_cluster.utils.utils.wait_for_condition")
def test_connect_to_node(
    mock_wait_for_condition, mock_run, mock_sleep, mock_cluster_spec
):
    """Test connect_to_node() in RayCluster"""
    ray_cluster = RayCluster()
    mock_run.return_value = mock.MagicMock(stdout="1 node_name")
    mock_wait_for_condition.return_value = "done"
    mock_sleep.return_value = "1s"
    ray_cluster.connect_to_node(node_address="mock_address")
    assert mock_wait_for_condition.called
    assert mock_run.called



import os
from pathlib import Path
from kfp.dsl import Output, Input, Artifact, component, Model

_COMPONENTS_CONTAINER_REGISTRY = os.environ.get(
    "COMPONENTS_CONTAINER_REGISTRY",
    "europe-west2-docker.pkg.dev/dt-pc-dev/dt-pc-dev-components-repository",
)
_IMAGE_NAME = str(Path(__file__).parents[1].name)
_COMPONENTS_IMAGE_TAG = os.environ.get("COMPONENTS_IMAGE_TAG", "latest")


@component(
    base_image="python:3.9-slim-buster",
    target_image=f"{_COMPONENTS_CONTAINER_REGISTRY}/{_IMAGE_NAME}:{_COMPONENTS_IMAGE_TAG}",  # noqa: E501
    output_component_file=os.path.splitext(__file__)[0] + ".yaml",
)
def upload_pytorch_model(
    project_id: str,
    location: str,
    gcs_model_path: Input[Artifact],
    local_model_file: str,
    vertex_model: Output[Model],
    model_display_name: str,
    handler: str,
    model_name: str = "model",
    version: str = "v1.0",
    extra_files_to_ignore: list = [],
    extra_files_staging_path: str = None,
    extra_prediction_files: str = None,
    requirements: str = None,
    custom_tokenizer_path: str = None,
    custom_tokenizer_name: str = None,
    custom_container: bool = False,
    serving_container_image_uri: str = "europe-docker.pkg.dev/vertex-ai/prediction/pytorch-cpu.1-12:latest",
    health_route: str = "/ping",
    predict_route: str = "/predictions/model",
    serving_container_ports: list = [8080],
    serving_environment_vars: dict = {"MODEL_NAME": "model"},
    serving_container_args: list = None,
    serving_container_command: list = None,
    model_description: str = None,
    parent_model: str = None,
    is_default_version: bool = True,
    version_description: str = None,
):
    """
    This component receives as an input a trained PyTorch model artifact stored in GCS and
    outputs a Vertex Model uploaded in Vertex Model Registry and ready to serve predictions
    in an Endpoint or via Batch Predictions

    Optionally, this component also accepts a custom tokenizer as input, if this tokenizer
    is used to serve predictions.

    The logic of the component is as follows:
    1. Stage all required files (saved in GCS) locally in the component execution environment
    2. Use `torch-model-archiver` to package the model into the required format for deployment
        and with the requried files included in the package. This returns a <model_name>.mar file.
    3. Upload the packaged model (`<model_name>.mar`) back to GCS
    4. Upload the packaged model to the Vertex Model Registry together with either a pre-built
        container or a custom container

    For more information on torch-model-archiver refer to
    https://github.com/pytorch/serve/blob/master/model-archiver/README.md

    Both the pre-built container and any custom container you build for predictions run with
    TorchServe https://github.com/pytorch/serve TorchServe offers a built-in prediction handler
    routine to serve your model. However, if your model requires a custom prediction routine
    you will need to provide a custom handler as input, and this handler will be part of the
    packaged .mar file. Note that if this prediction routine requires extra packages not
    included in the base pre-built container image, you will need to use a custom container
    for your predictions. Most importantly, this custom container contains a file called
    config.properties, which should have a new entry called install_py_dep_per_model=true.

    Args:
        project_id: str - GCP project ID
        location: str - GCP location
        gcs_model_path: Artifact - KFP artifact that includes the path to the model artifact in the uri
        local_model_file: str - File name of the model artifact (ie: pytorch_model.bin)
        model_display_name: str - Display name for Vertex Model in the Model Registry
        handler: str - Name of custom handler file used for serving. If using the default
            TorchServe one provide its name
        model_name: str - Model name used for TorchServe (recommended to leave as default 'model')
        version: str - Model version used for TorchServe (recommended to leave as default 'v1.0')
        extra_files_to_ignore: list - List of files included in the GCS model folder that you don't
            wish to include in the .mar package
        extra_files_staging_path: str - Path to GCS folder that contains additional files that must
            be included in the .mar package
        extra_prediction_files: str - Name of additional files that must be included in the .mar
            package for serving predictions
        requirements: str - Additional packages required by the model for serving predictions
        custom_tokenizer_path: str - Path to GCS file where the custom tokenizer is located.
            Leave blank if your model uses a default tokenizer
        custom_tokenizer_name: str - Name of custom tokenizer. Leave blank if your model uses
            a default tokenizer
        custom_container: bool - Whether to use a custom container or the pre-built container.
            If using a custom, the arguments below will be used
        serving_container_image_uri: str - Path to container image used for serving.
            Uses the pre-built Vertex container by default
        health_route: str - Overwrites the default health route in the serving container.
            Must be set to "/ping". Only used is custom_container=True
        predict_route: str - Overwrites the default prediction route in the serving container.
            Must be set to "/predictions/model"". Only used is custom_container=True
        serving_container_ports: list - Overwrites the default ports in the serving container.
            Only used is custom_container=True
        serving_environment_vars: dict - Overwrites the default environment variables in the serving container.
            Only used is custom_container=True
        serving_container_args: list - Overwrites the CMD of the Vertex pre-built container.
            Only used is custom_container=False
        serving_container_command: list - Overwrites the ENTRYPOINT of the Vertex pre-built container.
            Only used is custom_container=False
        model_description: str - Model description used in the Model Registry
        parent_model: str - The resource name or model ID of an existing model that the
            newly-uploaded model will be a version of. Only set this field when uploading a
            new version of an existing model.
        is_default_version: bool - When set to True, the newly uploaded model version will
            automatically have alias "default" included. Subsequent uses of this model without
            a version specified will use this "default" version. When set to False, the "default"
            alias will not be moved. Actions targeting the newly-uploaded model version will
            need to specifically reference this version by ID or alias. New model uploads,
            i.e. version 1, will always be "default" aliased.,
        version_description: str - The description of the model version being uploaded.,

    Returns:
        vertex_model: Model - Vertex model in Model Registry. Metadata includes:
            DisplayName - Display name shown in Vertex Model Registry
            VertexModelID - Full path to Vertex Model Registry ID (<project>/<location>/model/ID)
    """

    from google.cloud import aiplatform

    from upload_pytorch_model.utils.gcs_helper import split_bucket_object_file
    from upload_pytorch_model.utils.model_archiver import create_model_archiver

    # Run checks to ensure variables are valid
    if not custom_container and model_name != "model":
        print("Wrong model_name specified")
        raise NameError(
            f"""
        When using the pre-built PyTorch container the model_name must be 'model' but you selected {model_name}
        """
        )

    if not custom_container and requirements:
        print("Custom container not being used with custom packages")
        raise TypeError(
            f"""
        If your model requires extra packages you must use a custom container for serving predictions
        """
        )

    # Extract the correct model folder
    gcs_model_full_path = gcs_model_path.uri
    model_bucket_name, model_object_name, _ = split_bucket_object_file(
        path=gcs_model_full_path, file_name=local_model_file
    )
    gcs_model_path = f"gs://{model_bucket_name}/{model_object_name}"

    # Package model into .mar
    create_model_archiver(
        gcs_model_path=gcs_model_path,
        model_name=model_name,
        version=version,
        local_model_file=local_model_file,
        extra_files_staging_path=extra_files_staging_path,
        extra_files_to_ignore=extra_files_to_ignore,
        extra_prediction_files=extra_prediction_files,
        handler=handler,
        requirements=requirements,
        custom_tokenizer_path=custom_tokenizer_path,
        custom_tokenizer_name=custom_tokenizer_name,
    )

    # Define container serving variables
    serving_container_predict_route = predict_route if custom_container else None
    serving_container_environment_variables = (
        serving_environment_vars if custom_container else None
    )
    serving_container_health_route = health_route if custom_container else None
    serving_container_ports = serving_container_ports if custom_container else None

    # Upload to Vertex Model Registry
    aiplatform.init(project=project_id, location=location)

    model = aiplatform.Model.upload(
        display_name=model_display_name,
        description=model_description,
        serving_container_image_uri=serving_container_image_uri,
        serving_container_predict_route=serving_container_predict_route,
        serving_container_health_route=serving_container_health_route,
        serving_container_ports=serving_container_ports,
        serving_container_command=serving_container_command,
        serving_container_environment_variables=serving_container_environment_variables,
        serving_container_args=serving_container_args,
        artifact_uri=gcs_model_path,
        parent_model=parent_model,
        is_default_version=is_default_version,
        version_description=version_description,
    )

    model.wait()

    vertex_model.metadata["DisplayName"] = model.display_name
    vertex_model.metadata["VertexModelID"] = model.resource_name
    vertex_model.uri = model.resource_name
